# Copyright 2020 Manuel Genovés
# Copyright 2022 Mufeed Ali
# Copyright 2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

# Code modified from Apostrophe
# https://gitlab.gnome.org/World/apostrophe/-/blob/main/apostrophe/theme_switcher.py

from gi.repository import Adw, Gio, GObject, Gtk

from dialect.define import RES_PATH
from dialect.settings import Settings


@Gtk.Template(resource_path=f"{RES_PATH}/widgets/theme_switcher.ui")
class ThemeSwitcher(Gtk.Box):
    __gtype_name__ = "ThemeSwitcher"

    # Properties
    show_system: bool = GObject.property(type=bool, default=True)  # type: ignore
    color_scheme: str = "light"

    # Child widgets
    system: Gtk.CheckButton = Gtk.Template.Child()  # type: ignore
    light: Gtk.CheckButton = Gtk.Template.Child()  # type: ignore
    dark: Gtk.CheckButton = Gtk.Template.Child()  # type: ignore

    @GObject.Property(type=str)
    def selected_color_scheme(self) -> str:  # type: ignore
        """Read-write integer property."""

        return self.color_scheme

    @selected_color_scheme.setter
    def selected_color_scheme(self, color_scheme: str):
        self.color_scheme = color_scheme

        if color_scheme == "auto":
            self.system.props.active = True
            self.style_manager.props.color_scheme = Adw.ColorScheme.PREFER_LIGHT
        if color_scheme == "light":
            self.light.props.active = True
            self.style_manager.props.color_scheme = Adw.ColorScheme.FORCE_LIGHT
        if color_scheme == "dark":
            self.dark.props.active = True
            self.style_manager.props.color_scheme = Adw.ColorScheme.FORCE_DARK

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.style_manager = Adw.StyleManager.get_default()

        self.color_scheme = Settings.get().color_scheme

        Settings.get().bind("color-scheme", self, "selected_color_scheme", Gio.SettingsBindFlags.DEFAULT)

        self.style_manager.bind_property(
            "system-supports-color-schemes", self, "show_system", GObject.BindingFlags.SYNC_CREATE
        )

    @Gtk.Template.Callback()
    def _on_color_scheme_changed(self, _widget, _paramspec):
        """Called on (self.system, self.light, self.dark)::notify::active signal"""
        if self.system.props.active:
            self.selected_color_scheme = "auto"
        if self.light.props.active:
            self.selected_color_scheme = "light"
        if self.dark.props.active:
            self.selected_color_scheme = "dark"



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from dialect.providers.base import (
    ProviderCapability,
    ProviderFeature,
    Translation,
)
from dialect.providers.errors import (
    APIKeyInvalid,
    APIKeyRequired,
    BatchSizeExceeded,
    CharactersLimitExceeded,
    InvalidLangCode,
    UnexpectedError,
)
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "libretranslate"
    prettyname = "LibreTranslate"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.INSTANCES | ProviderFeature.DETECTION

    defaults = {
        "instance_url": "lt.dialectapp.org",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 0

    @property
    def frontend_settings_url(self):
        return self.format_url(self.instance_url, "/frontend/settings")

    @property
    def detect_url(self):
        return self.format_url(self.instance_url, "/detect")

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, "/languages")

    @property
    def suggest_url(self):
        return self.format_url(self.instance_url, "/suggest")

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, "/translate")

    async def validate_instance(self, url):
        response = await self.get(self.format_url(url, "/spec"), check_common=False)
        valid = False

        try:
            valid = response["info"]["title"] == "LibreTranslate"
        except:  # noqa
            pass

        return valid

    async def validate_api_key(self, key):
        # Form data
        data = {
            "q": "hello",
            "api_key": key,
        }

        try:
            response = await self.post(self.detect_url, data, form=True)
            return "confidence" in response[0]
        except (APIKeyInvalid, APIKeyRequired):
            return False
        except Exception:
            raise

    async def init_trans(self):
        languages = await self.get(self.lang_url)
        settings = await self.get(self.frontend_settings_url)

        try:
            for lang in languages:
                self.add_lang(lang["code"], lang["name"])

            if settings.get("suggestions", False):
                self.features ^= ProviderFeature.SUGGESTIONS
            if settings.get("apiKeys", False):
                self.features ^= ProviderFeature.API_KEY
            if settings.get("keyRequired", False):
                self.features ^= ProviderFeature.API_KEY_REQUIRED

            self.chars_limit = int(settings.get("charLimit", 0))

        except Exception as exc:
            raise UnexpectedError from exc

    async def translate(self, request):
        src, dest = self.denormalize_lang(request.src, request.dest)

        # Request body
        data = {
            "q": request.text,
            "source": src,
            "target": dest,
        }
        if self.api_key and ProviderFeature.API_KEY in self.features:
            data["api_key"] = self.api_key

        # Do request
        response = await self.post(self.translate_url, data)
        try:
            detected = response.get("detectedLanguage", {}).get("language", None)
            return Translation(response["translatedText"], request, detected)
        except Exception as exc:
            raise UnexpectedError from exc

    async def suggest(self, text, src, dest, suggestion):
        src, dest = self.denormalize_lang(src, dest)

        # Form data
        data = {
            "q": text,
            "source": src,
            "target": dest,
            "s": suggestion,
        }
        if self.api_key and ProviderFeature.API_KEY in self.features:
            data["api_key"] = self.api_key

        # Do request
        response = await self.post(self.suggest_url, data, form=True)
        try:
            return response.get("success", False)
        except:  # noqa
            return False

    def check_known_errors(self, status, data):
        if not data:
            raise UnexpectedError("Response is empty!")

        if "error" in data:
            error = data["error"]

            if error == "Please contact the server operator to obtain an API key":
                raise APIKeyRequired(error)
            elif error == "Invalid API key":
                raise APIKeyInvalid(error)
            elif "is not supported" in error:
                raise InvalidLangCode(error)
            elif "exceeds text limit" in error:
                raise BatchSizeExceeded(error)
            elif "exceeds character limit" in error:
                raise CharactersLimitExceeded(error)
            else:
                raise UnexpectedError(error)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib
import logging
import pkgutil

from gi.repository import Gio, GObject

from dialect.providers import modules
from dialect.providers.base import (  # noqa
    BaseProvider,
    ProviderCapability,
    ProviderFeature,
    Translation,
    TranslationMistake,
    TranslationPronunciation,
    TranslationRequest,
)
from dialect.providers.errors import (  # noqa
    APIKeyInvalid,
    APIKeyRequired,
    BatchSizeExceeded,
    CharactersLimitExceeded,
    InvalidLangCode,
    ProviderError,
    RequestError,
    ServiceLimitReached,
    UnexpectedError,
)

MODULES: dict[str, type[BaseProvider]] = {}
TRANSLATORS: dict[str, type[BaseProvider]] = {}
TTS: dict[str, type[BaseProvider]] = {}
for _importer, modname, _ispkg in pkgutil.iter_modules(modules.__path__):
    try:
        modclass = importlib.import_module("dialect.providers.modules." + modname).Provider
        MODULES[modclass.name] = modclass
        if modclass.capabilities:
            if ProviderCapability.TRANSLATION in modclass.capabilities:
                TRANSLATORS[modclass.name] = modclass
            if ProviderCapability.TTS in modclass.capabilities:
                TTS[modclass.name] = modclass
    except Exception as exc:
        logging.warning(f"Could not load the {modname} provider: {exc}")


def check_translator_availability(provider_name: str) -> bool:
    if provider_name in TRANSLATORS:
        return True
    return False


def get_fallback_translator_name() -> str:
    if TRANSLATORS:
        return next(iter(TRANSLATORS))
    return ""


class ProviderObject(GObject.Object):
    __gtype_name__ = "ProviderObject"

    def __init__(self, p_class: type[BaseProvider] | None = None):
        super().__init__()

        self.p_class = p_class

    @GObject.Property(type=str)
    def name(self) -> str:
        if self.p_class is not None:
            return self.p_class.name
        else:
            return ""

    @GObject.Property(type=str)
    def prettyname(self) -> str:
        if self.p_class is not None:
            return self.p_class.prettyname
        else:
            return _("Disabled")


class ProvidersListModel(GObject.GObject, Gio.ListModel):
    __gtype_name__ = "ProvidersListModel"

    def __init__(self, p_type="", show_disabled=False):
        super().__init__()

        if p_type:  # If we want to get an specific provider type
            types = {"translators": TRANSLATORS, "tts": TTS}
            providers = types.get(p_type)
        else:  # Get all providers
            providers = MODULES

        providers = list(providers.values())  # type: ignore
        self.providers: list[ProviderObject] = []
        for provider in providers:
            self.providers.append(ProviderObject(provider))
        if show_disabled and self.providers:
            self.providers.insert(0, ProviderObject())

    def do_get_item(self, position: int) -> ProviderObject:
        return self.providers[position]

    def do_get_item_type(self):
        return ProviderObject

    def do_get_n_items(self) -> int:
        return len(self.providers)

    def get_index_by_name(self, name: str) -> int:
        for i, prov in enumerate(self.providers):
            if prov.name == name:
                return i
        return 0



# Copyright 2022 Mufeed Ali
# Copyright 2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging
from asyncio import sleep
from typing import Any

from gi.repository import GLib, Soup

from dialect.providers.base import BaseProvider
from dialect.providers.errors import RequestError
from dialect.session import Session


class SoupProvider(BaseProvider):
    """Base class for providers needing libsoup helpers"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.retry_errors: tuple[int] = tuple()
        """ Error codes that should be retried automatically """
        self.max_retries = 5
        """ Max number of tries """

    def encode_data(self, data: Any) -> GLib.Bytes | None:
        """
        Convert Python data to JSON and bytes.

        Args:
            data: Data to encode, anything json.dumps can handle.

        Returns:
            The GLib Bytes or None if something failed.
        """
        data_glib_bytes = None
        try:
            data_bytes = json.dumps(data).encode("utf-8")
            data_glib_bytes = GLib.Bytes.new(data_bytes)
        except Exception as exc:
            logging.warning(exc)
        return data_glib_bytes

    def create_message(
        self, method: str, url: str, data: Any = {}, headers: dict = {}, form: bool = False
    ) -> Soup.Message:
        """
        Create a Soup's message.

        Encodes data and adds it to the message as the request body.

        Args:
            method: HTTP method of the message.
            url: Url of the message.
            data: Request body or form data.
            headers: HTTP headers of the message.
            form: If the data should be encoded as ``application/x-www-form-urlencoded``.

        Returns:
            The Soup Message for the given parameters.
        """

        if form and data:
            form_data = Soup.form_encode_hash(data)
            message = Soup.Message.new_from_encoded_form(method, url, form_data)
        else:
            message = Soup.Message.new(method, url)

        if message:
            if data and not form:
                data = self.encode_data(data)
                message.set_request_body_from_bytes("application/json", data)
            if headers:
                for name, value in headers.items():
                    message.get_request_headers().append(name, value)
            if "User-Agent" not in headers:
                message.get_request_headers().append("User-Agent", "Dialect App")

        return message  # type: ignore

    async def send_and_read(self, message: Soup.Message) -> bytes | None:
        """
        Helper method for Soup's send_and_read_async.

        Args:
            message: Message to send.

        Returns:
            The bytes of the response or None.
        """
        response: GLib.Bytes = await Session.get().send_and_read_async(message, 0)  # type: ignore
        return response.get_data()

    async def send_and_read_json(self, message: Soup.Message) -> Any:
        """
        Like ``SoupProvider.send_and_read`` but returns JSON parsed.

        Args:
            message: Message to send.

        Returns:
            The JSON of the response deserialized to a python object.
        """
        response = await self.send_and_read(message)
        return json.loads(response) if response else {}

    def check_known_errors(self, status: Soup.Status, data: Any) -> None:
        """
        Checks data for possible response errors and raises appropriated exceptions.

        This should be implemented by subclases.

        Args:
            status: HTTP status.
            data: Response body data.
        """

    async def send_and_read_and_process(
        self,
        message: Soup.Message,
        check_common: bool = True,
        return_json: bool = True,
    ) -> Any:
        """
        Helper mixing ``SoupProvider.send_and_read``, ``SoupProvider.send_and_read_json``
        and ``SoupProvider.check_known_errors``.

        Converts `GLib.Error` to `RequestError`.

        It also handles retries for status codes listen in ``self.retry_errors``.

        Args:
            message: Message to send.
            check_common: If response data should be checked for errors using check_known_errors.
            return_json: If the response should be parsed as JSON.

        Returns:
            The JSON deserialized to a python object or bytes if ``json`` is ``False``.
        """

        async def send_and_read() -> Any:
            if return_json:
                return await self.send_and_read_json(message)
            else:
                return await self.send_and_read(message)

        try:
            response = await send_and_read()

            # Do retries with exponential backoff for errors
            if message.get_status() in self.retry_errors:
                delay = 1

                for _ in range(self.max_retries):
                    await sleep(delay)
                    response = await send_and_read()

                    if message.get_status() in self.retry_errors:
                        delay *= 2
                    else:
                        break

            if check_common:
                self.check_known_errors(message.get_status(), response)

            return response
        except GLib.Error as exc:
            raise RequestError(exc.message)

    async def request(
        self,
        method: str,
        url: str,
        data: Any = {},
        headers: dict = {},
        form: bool = False,
        check_common: bool = True,
        return_json: bool = True,
    ) -> Any:
        """
        Helper for regular HTTP request.

        Args:
            method: HTTP method of the request.
            url: Url of the request.
            data: Request body or form data.
            headers: HTTP headers of the message.
            form: If the data should be encoded as a form.
            check_common: If response data should be checked for errors using check_known_errors.
            return_json: If the response should be parsed as JSON.

        Returns:
            The JSON deserialized to a python object or bytes if ``json`` is ``False``.
        """
        message = self.create_message(method, url, data, headers, form)
        return await self.send_and_read_and_process(message, check_common, return_json)

    async def get(
        self,
        url: str,
        headers: dict = {},
        check_common: bool = True,
        return_json: bool = True,
    ) -> Any:
        """
        Helper for GET HTTP request.

        Args:
            url: Url of the request.
            headers: HTTP headers of the message.
            check_common: If response data should be checked for errors using check_known_errors.
            return_json: If the response should be parsed as JSON.

        Returns:
            The JSON deserialized to a python object or bytes if ``json`` is ``False``.
        """
        return await self.request("GET", url, headers=headers, check_common=check_common, return_json=return_json)

    async def post(
        self,
        url: str,
        data: Any = {},
        headers: dict = {},
        form: bool = False,
        check_common: bool = True,
        return_json: bool = True,
    ) -> Any:
        """
        Helper for POST HTTP request.

        Args:
            url: Url of the request.
            data: Request body or form data.
            headers: HTTP headers of the message.
            form: If the data should be encoded as a form.
            check_common: If response data should be checked for errors using check_known_errors.
            return_json: If the response should be parsed as JSON.

        Returns:
            The JSON deserialized to a python object or bytes if ``json`` is ``False``.
        """
        return await self.request("POST", url, data, headers, form, check_common, return_json)



# Copyright 2024 Mufeed Ali
# Copyright 2024 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later


def find_item_match(list1: list[str], list2: list[str]) -> str | None:
    """
    Get the first occurrence in two lists.

    Args:
        list1: List to iterate
        list2: List to check against
    """
    set_to_check = set(list2)
    return next((i for i in list1 if i in set_to_check), None)


def first_exclude(list_: list[str], exclude: str) -> str | None:
    """
    Get the first item that is not excluded.

    Args:
        list_: List of items
        exclude: Item to ignore
    """
    return next((x for x in list_ if x != exclude), None)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import urllib.parse
from dataclasses import dataclass, field
from enum import Enum, Flag, auto
from typing import IO

from dialect.define import LANG_ALIASES
from dialect.languages import get_lang_name
from dialect.providers.settings import ProviderDefaults, ProviderSettings


class ProviderCapability(Flag):
    TRANSLATION = auto()
    """ If it provides translation """
    TTS = auto()
    """ If it provides text-to-speech """
    DEFINITIONS = auto()
    """ If it provides dictionary definitions """


class ProviderFeature(Flag):
    NONE = auto()
    """ Provider has no features """
    INSTANCES = auto()
    """ If it supports changing the instance url """
    API_KEY = auto()
    """ If the api key is supported but not necessary """
    API_KEY_REQUIRED = auto()
    """ If the api key is required for the provider to work """
    API_KEY_USAGE = auto()
    """ If the service reports api usage """
    DETECTION = auto()
    """ If it supports detecting text language (Auto translation) """
    MISTAKES = auto()
    """ If it supports showing translation mistakes """
    PRONUNCIATION = auto()
    """ If it supports showing translation pronunciation """
    SUGGESTIONS = auto()
    """ If it supports sending translation suggestions to the service """


class ProviderLangModel(Enum):
    STATIC = auto()
    """
    The provider populate its `src_languages` and `dest_languages` properties.
    The `cmp_langs` method will be used to decide if one code can be translated to another.
    """
    DYNAMIC = auto()
    """
    The provider only populate its `src_languages` property.
    The `dest_langs_for` method will be used to get possible destination codes for a code.
    """


class ProviderLangComparison(Enum):
    PLAIN = auto()
    """Perform a simple language codes comparision. a == b"""
    DEEP = auto()
    """Check for ISO 3166-1 and 15924 codes, and compare if base languages """


@dataclass
class TranslationRequest:
    text: str
    src: str
    dest: str


@dataclass
class TranslationMistake:
    markup: str
    text: str


@dataclass
class TranslationPronunciation:
    src: str | None
    dest: str | None


@dataclass
class Translation:
    text: str
    original: TranslationRequest
    detected: str | None = None
    mistakes: TranslationMistake | None = None
    pronunciation: TranslationPronunciation = field(default_factory=lambda: TranslationPronunciation(None, None))


class BaseProvider:
    name = ""
    """ Module name for code use, like settings storing """
    prettyname = ""
    """ Module name for UI display """
    capabilities: ProviderCapability | None = None
    """ Provider capabilities, translation, tts, etc """
    features: ProviderFeature = ProviderFeature.NONE
    """ Provider features """
    lang_model: ProviderLangModel = ProviderLangModel.STATIC
    """ Translation language model """
    lang_comp: ProviderLangComparison = ProviderLangComparison.PLAIN
    """ Define behavior of default `cmp_langs` method """

    defaults: ProviderDefaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }
    """ Default provider settings """

    def __init__(self):
        self.src_languages: list[str] = []
        """ Source languages available for translating """
        self.dest_languages: list[str] = []
        """ Destination languages available for translating """
        self.tts_languages: list[str] = []
        """ Languages available for TTS """
        self._nonstandard_langs: dict[str, str] = {}
        """ Mapping of lang codes that differ with Dialect ones """
        self._languages_names: dict[str, str] = {}
        """ Names of languages provided by the service """

        self.chars_limit: int = -1
        """ Translation char limit """

        self.history: list[Translation] = []
        """ Here we save the translation history """

        # GSettings
        self.settings = ProviderSettings(self.name, self.defaults)

    """
    Providers API methods
    """

    async def validate_instance(self, url: str) -> bool:
        """
        Validate an instance of the provider.

        Args:
            url: The instance URL to test, only hostname and tld, e.g. ``libretranslate.com``, ``localhost``.

        Returns:
            If the URL is a valid instance of the provider ot not.
        """
        raise NotImplementedError()

    async def validate_api_key(self, key: str) -> bool:
        """
        Validate an API key.

        Args:
            key: The API key to validate.

        Returns:
            If the API key is valid or not.
        """
        raise NotImplementedError()

    async def init_trans(self) -> None:
        """Initializes the provider translation capabilities."""
        raise NotImplementedError()

    async def init_tts(self) -> None:
        """Initializes the provider text-to-speech capabilities."""
        raise NotImplementedError()

    async def translate(self, request: TranslationRequest) -> Translation:
        """
        Translates text in the provider.

        Providers are expected to use ``BaseProvider.denormalize_lang`` because
        ``request`` will use normalized lang codes.

        Args:
            request: The translation request.

        Returns:
            A new translation object.
        """
        raise NotImplementedError()

    async def suggest(self, text: str, src: str, dest: str, suggestion: str) -> bool:
        """
        Sends a translation suggestion to the provider.

        Providers are expected to use ``BaseProvider.denormalize_lang`` because
        ``src`` and ``dest`` will use normalized lang codes.

        Args:
            text: Original text without translation.
            src: The lang code of the original text.
            dest: The lang code of the translated text.
            suggestion: Suggested translation for text.

        Returns:
            If the suggestion was successful or not.
        """
        raise NotImplementedError()

    async def speech(self, text: str, language: str) -> IO:
        """
        Generate speech audio from text.

        Providers are expected to use ``BaseProvider.denormalize_lang`` because
        ``language`` will use normalized lang codes.

        Args:
            text: Text to generate speech from.
            language: The lang code of text.

        Returns:
            The file object with the speech audio written.
        """
        raise NotImplementedError()

    async def api_char_usage(self) -> tuple[int, int]:
        """
        Retrieves the API usage status.

        Returns:
            The current usage and limit.
        """
        raise NotImplementedError()

    def cmp_langs(self, a: str, b: str) -> bool:
        """
        Compare two language codes.

        It assumes that the codes have been normalized by ``BaseProvider.normalize_lang_code``.

        This method exists so providers can add additional comparison logic.
        Default behavior depends on `self.lang_comp` value.

        Args:
            a: First lang to compare.
            b: Second lang to compare.

        Returns:
            Whether both languages are equals in some way or not.
        """

        # Early return if both langs are just the same
        if a == b:
            return True

        # Plain comparison
        if self.lang_comp == ProviderLangComparison.PLAIN:
            return a == b

        # Split lang code to separate possible country/script code
        a_codes = a.split("-")
        b_codes = b.split("-")

        if a_codes[0] == b_codes[0]:  # Check base codes
            return True

        return False


    def dest_langs_for(self, code: str) -> list[str]:
        """
        Get the available destination languages for a source language.

        Returns:
            The codes of available languages.
        """
        raise NotImplementedError()

    @property
    def lang_aliases(self) -> dict[str, str]:
        """
        Mapping of Dialect/CLDR's lang codes to the provider ones.

        Some providers might use different lang codes from the ones used by Dialect.

        This dict is used by ``BaseProvider.add_lang`` so lang codes can later be denormalized with
        ``BaseProvider.denormalize_lang``.

        Codes must be formatted with the criteria from ``BaseProvider.normalize_lang_code``, because this value would
        be used by ``BaseProvider.add_lang`` after normalization.

        Check ``dialect.define.LANG_ALIASES`` for reference mappings.
        """
        return {}

    """
    Provider features helpers
    """

    @property
    def supports_instances(self) -> bool:
        return ProviderFeature.INSTANCES in self.features

    @property
    def supports_api_key(self) -> bool:
        return ProviderFeature.API_KEY in self.features

    @property
    def api_key_required(self) -> bool:
        return ProviderFeature.API_KEY_REQUIRED in self.features

    @property
    def supports_api_usage(self) -> bool:
        return ProviderFeature.API_KEY_USAGE in self.features

    @property
    def supports_detection(self) -> bool:
        return ProviderFeature.DETECTION in self.features

    @property
    def supports_mistakes(self) -> bool:
        return ProviderFeature.MISTAKES in self.features

    @property
    def supports_pronunciation(self) -> bool:
        return ProviderFeature.PRONUNCIATION in self.features

    @property
    def supports_suggestions(self) -> bool:
        return ProviderFeature.SUGGESTIONS in self.features

    """
    Provider settings helpers and properties
    """

    @property
    def instance_url(self) -> str:
        """Instance url saved on settings"""
        return self.settings.instance_url

    @instance_url.setter
    def instance_url(self, url: str):
        self.settings.instance_url = url

    def reset_instance_url(self):
        """Resets saved instance url"""
        self.instance_url = ""

    @property
    def api_key(self) -> str:
        """API key saved on settings"""
        return self.settings.api_key

    @api_key.setter
    def api_key(self, api_key: str):
        self.settings.api_key = api_key

    def reset_api_key(self):
        """Resets saved API key."""
        self.api_key = ""

    @property
    def recent_src_langs(self) -> list[str]:
        """Saved recent source langs of the user"""
        return self.settings.src_langs

    @recent_src_langs.setter
    def recent_src_langs(self, src_langs: list[str]):
        self.settings.src_langs = src_langs

    def reset_src_langs(self):
        """Reset saved recent user source langs"""
        self.recent_src_langs = []

    @property
    def recent_dest_langs(self) -> list[str]:
        """Saved recent destination langs of the user"""
        return self.settings.dest_langs

    @recent_dest_langs.setter
    def recent_dest_langs(self, dest_langs: list[str]):
        self.settings.dest_langs = dest_langs

    def reset_dest_langs(self):
        """Reset saved recent user destination langs"""
        self.recent_dest_langs = []

    """
    General provider helpers
    """

    @staticmethod
    def format_url(url: str, path: str = "", params: dict = {}, http: bool = False) -> str:
        """
        Compose a HTTP url with the given pieces.

        If url is "localhost", ``http`` is ignored and HTTP protocol is forced.

        Args:
            url: Base url, hostname and tld.
            path: Path of the url.
            params: Params to populate a url query.
            http: If HTTP should be used instead of HTTPS.

        Returns:
            The new formatted URL.
        """

        if not path.startswith("/"):
            path = "/" + path

        protocol = "https://"
        if url.startswith("localhost:") or http:
            protocol = "http://"

        params_str = urllib.parse.urlencode(params)
        if params_str:
            params_str = "?" + params_str

        return protocol + url + path + params_str

    def normalize_lang_code(self, code: str) -> str:
        """
        Normalice a language code with Dialect's criteria.

        This method also maps to lang codes aliases using ``BaseProvider.lang_aliases`` and
        ``dialect.define.LANG_ALIASES``.

        Criteria:
            - Codes must be lowercase, e.g. ES => es
            - Codes can have a second code delimited by a hyphen, e.g. zh_CN => zh-CN
            - If second code is two chars long it's considered a country code and must be uppercase, e.g. zh-cn => zh-CN
            - If second code is four chars long it's considered a script code and must be capitalized,
            e.g. zh-HANS => zh-Hans

        Args:
            code: Language ISO code.

        Returns:
            The normalize language code.
        """
        code = code.replace("_", "-").lower()  # Normalize separator
        codes = code.split("-")

        if len(codes) == 2:  # Code contain a script or country code
            if len(codes[1]) == 4:  # ISO 15924 (script)
                codes[1] = codes[1].capitalize()

            elif len(codes[1]) == 2:  # ISO 3166-1 (country)
                codes[1] = codes[1].upper()

            code = "-".join(codes)

        aliases = {**LANG_ALIASES, **self.lang_aliases}
        if code in aliases:
            code = aliases[code]

        return code

    def add_lang(
        self,
        original_code: str,
        name: str | None = None,
        trans_src: bool = True,
        trans_dest: bool = True,
        tts: bool = False,
    ) -> None:
        """
        Register lang supported by the provider.

        Lang codes are normalized and saved for latter denormalization using
        ``BaseProvider.denormalize_lang``.

        Args:
            original_code: Lang code to add
            name: Language name to fallback in case Dialect doesn't provide one
            trans_src: Add language as supported for translation as src
            trans_dest: Add language as supported for translation as dest
            tts: Add language as supported for text-to-speech
        """

        code = self.normalize_lang_code(original_code)  # Get normalized lang code

        if trans_src:  # Add lang to supported languages list
            self.src_languages.append(code)
        if trans_dest:
            self.dest_languages.append(code)
        if tts:  # Add lang to supported TTS languages list
            self.tts_languages.append(code)

        if code != original_code and code not in self._nonstandard_langs:
            # Save a divergent lang code for later denormalization
            self._nonstandard_langs[code] = original_code

        if name is not None and code not in self._languages_names:
            # Save name provided by the service
            self._languages_names[code] = name

    def denormalize_lang(self, *codes: str) -> tuple[str, ...]:
        """
        Get denormalized lang code if available.

        Args:
            *codes: Lang codes to denormalize

        Returns:
            The same amount of given codes but denormalized.
        """

        result = []
        for code in codes:
            result.append(self._nonstandard_langs.get(code, code))
        return tuple(result)

    def get_lang_name(self, code: str) -> str | None:
        """
        Get a localized language name.

        Fallback to a name provided by the provider if available or ultimately
        just the code.

        Args:
            code: Language to get a name for

        Returns:
            The language name.
        """
        name = get_lang_name(code)  # Try getting translated name from Dialect

        if name is None:  # Get name from provider if available
            return self._languages_names.get(code, code)

        return name



# Copyright 2023 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from bs4 import BeautifulSoup, Tag

from dialect.providers.base import ProviderCapability, ProviderFeature, Translation, TranslationPronunciation
from dialect.providers.errors import ProviderError, UnexpectedError
from dialect.providers.soup import SoupProvider


class Provider(SoupProvider):
    name = "bing"
    prettyname = "Bing"

    capabilities = ProviderCapability.TRANSLATION
    features = ProviderFeature.DETECTION | ProviderFeature.PRONUNCIATION

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _headers = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)", "Accept": "*/*"}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 1000  # Web UI limit

        # Session vars
        self._key = ""
        self._token = ""
        self._ig = ""
        self._iid = ""
        self._count = 1

    @property
    def html_url(self):
        return self.format_url("www.bing.com", "/translator")

    @property
    def translate_url(self):
        params = {
            "isVertical": "1",
            "": "",
            "IG": self._ig,
            "IID": f"{self._iid}.{self._count}",
        }
        return self.format_url("www.bing.com", "/ttranslatev3", params)

    async def init_trans(self):
        response = await self.get(self.html_url, self._headers, check_common=False, return_json=False)

        if response:
            try:
                soup = BeautifulSoup(response, "html.parser")

                # Get Langs
                langs = soup.find("optgroup", {"id": "t_tgtAllLang"})
                if isinstance(langs, Tag):
                    for child in langs.findChildren():
                        if child.name == "option":
                            self.add_lang(child["value"], child.contents[0])

                # Get IID
                iid = soup.find("div", {"id": "rich_tta"})
                if isinstance(iid, Tag):
                    self._iid = iid["data-iid"]

                # Decode response bytes
                text = response.decode("utf-8")

                # Look for abuse prevention data
                params = re.findall(r"var params_AbusePreventionHelper = \[(.*?)\];", text)[0]  # noqa
                abuse_params = params.replace('"', "").split(",")
                self._key = abuse_params[0]
                self._token = abuse_params[1]

                # Look for IG
                self._ig = re.findall('IG:"(.*?)",', text)[0]

            except Exception as exc:
                raise UnexpectedError("Failed parsing HTML from bing.com") from exc

        else:
            raise UnexpectedError("Could not get HTML from bing.com")

    async def translate(self, request):
        src, dest = self.denormalize_lang(request.src, request.dest)

        # Increment requests count
        self._count += 1

        # Form data
        data = {
            "fromLang": "auto-detect" if src == "auto" else src,
            "text": request.text,
            "to": dest,
            "token": self._token,
            "key": self._key,
        }

        # Do request
        response = await self.post(self.translate_url, data, self._headers, True)

        try:
            data = response[0]
            detected = None
            pronunciation = None

            if "translations" in data:
                if "detectedLanguage" in data:
                    detected = data["detectedLanguage"]["language"]

                if "transliteration" in data["translations"][0]:
                    pronunciation = data["translations"][0]["transliteration"]["text"]

                return Translation(
                    data["translations"][0]["text"],
                    request,
                    detected=detected,
                    pronunciation=TranslationPronunciation(None, pronunciation),
                )
            else:
                raise UnexpectedError("Unexpected translation response")

        except Exception as exc:
            raise UnexpectedError from exc

    def check_known_errors(self, status, data):
        if not data:
            raise UnexpectedError("Response is empty!")

        if "errorMessage" in data:
            error = data["errorMessage"]
            code = data["statusCode"]

            match code:
                case _:
                    raise ProviderError(error)

        return None



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import html
import json
import random
import re
from tempfile import NamedTemporaryFile

from gtts import gTTS, lang

from dialect.providers.base import (
    ProviderCapability,
    ProviderFeature,
    Translation,
    TranslationMistake,
    TranslationPronunciation,
)
from dialect.providers.errors import UnexpectedError
from dialect.providers.local import LocalProvider
from dialect.providers.soup import SoupProvider

RPC_ID = "MkEWBc"

# Predefined URLs used to make google translate requests.
TRANSLATE_RPC = "{host}/_/TranslateWebserverUi/data/batchexecute"

TRANSLATE_TLDS = (
    ".ac",
    ".ad",
    ".ae",
    ".al",
    ".am",
    ".as",
    ".at",
    ".az",
    ".ba",
    ".be",
    ".bf",
    ".bg",
    ".bi",
    ".bj",
    ".bs",
    ".bt",
    ".by",
    ".ca",
    ".cat",
    ".cc",
    ".cd",
    ".cf",
    ".cg",
    ".ch",
    ".ci",
    ".cl",
    ".cm",
    ".cn",
    ".co.ao",
    ".co.bw",
    ".co.ck",
    ".co.cr",
    ".co.id",
    ".co.il",
    ".co.in",
    ".co.jp",
    ".co.ke",
    ".co.kr",
    ".co.ls",
    ".co.ma",
    ".co.mz",
    ".co.nz",
    ".co.th",
    ".co.tz",
    ".co.ug",
    ".co.uk",
    ".co.uz",
    ".co.ve",
    ".co.vi",
    ".co.za",
    ".co.zm",
    ".co.zw",
    ".com.af",
    ".com.ag",
    ".com.ai",
    ".com.ar",
    ".com.au",
    ".com.bd",
    ".com.bh",
    ".com.bn",
    ".com.bo",
    ".com.br",
    ".com.bz",
    ".com.co",
    ".com.cu",
    ".com.cy",
    ".com.do",
    ".com.ec",
    ".com.eg",
    ".com.et",
    ".com.fj",
    ".com.gh",
    ".com.gi",
    ".com.gt",
    ".com.hk",
    ".com.jm",
    ".com.kh",
    ".com.kw",
    ".com.lb",
    ".com.ly",
    ".com.mm",
    ".com.mt",
    ".com.mx",
    ".com.my",
    ".com.na",
    ".com.ng",
    ".com.ni",
    ".com.np",
    ".com.om",
    ".com.pa",
    ".com.pe",
    ".com.pg",
    ".com.ph",
    ".com.pk",
    ".com.pr",
    ".com.py",
    ".com.qa",
    ".com.sa",
    ".com.sb",
    ".com.sg",
    ".com.sl",
    ".com.sv",
    ".com.tj",
    ".com.tr",
    ".com.tw",
    ".com.ua",
    ".com.uy",
    ".com.vc",
    ".com.vn",
    ".com",
    ".cv",
    ".cz",
    ".de",
    ".dj",
    ".dk",
    ".dm",
    ".dz",
    ".ee",
    ".es",
    ".fi",
    ".fm",
    ".fr",
    ".ga",
    ".ge",
    ".gg",
    ".gl",
    ".gm",
    ".gp",
    ".gr",
    ".gy",
    ".hn",
    ".hr",
    ".ht",
    ".hu",
    ".ie",
    ".im",
    ".iq",
    ".is",
    ".it",
    ".je",
    ".jo",
    ".kg",
    ".ki",
    ".kz",
    ".la",
    ".li",
    ".lk",
    ".lt",
    ".lu",
    ".lv",
    ".md",
    ".me",
    ".mg",
    ".mk",
    ".ml",
    ".mn",
    ".ms",
    ".mu",
    ".mv",
    ".mw",
    ".ne",
    ".nl",
    ".no",
    ".nr",
    ".nu",
    ".pl",
    ".pn",
    ".ps",
    ".pt",
    ".ro",
    ".rs",
    ".ru",
    ".rw",
    ".sc",
    ".se",
    ".sh",
    ".si",
    ".sk",
    ".sm",
    ".sn",
    ".so",
    ".sr",
    ".st",
    ".td",
    ".tg",
    ".tk",
    ".tl",
    ".tm",
    ".tn",
    ".to",
    ".tt",
    ".us",
    ".vg",
    ".vu",
    ".ws",
)


class Provider(LocalProvider, SoupProvider):
    name = "google"
    prettyname = "Google"

    capabilities = ProviderCapability.TRANSLATION | ProviderCapability.TTS
    features = ProviderFeature.DETECTION | ProviderFeature.MISTAKES | ProviderFeature.PRONUNCIATION

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en"],
    }

    _headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)",
        "Referer": "https://translate.google.com",
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.chars_limit = 2000

    async def init_trans(self):
        langs_url = self.format_url(
            self._get_translate_host(".com"), "/translate_a/l", {"client": "t", "alpha": "true"}
        )
        response = await self.get(langs_url, self._headers, False)

        try:
            for code, name in response["tl"].items():
                self.add_lang(code, name)
        except Exception as exc:
            raise UnexpectedError from exc

    async def init_tts(self):
        for code in lang.tts_langs().keys():
            self.add_lang(code, trans_src=False, trans_dest=False, tts=True)

    @staticmethod
    def _build_rpc_request(text: str, src: str, dest: str):
        return json.dumps(
            [
                [
                    [
                        RPC_ID,
                        json.dumps([[text, src, dest, True], [None]], separators=(",", ":")),
                        None,
                        "generic",
                    ],
                ]
            ],
            separators=(",", ":"),
        )

    def _get_translate_host(self, tld: str | None = None):
        if not tld:
            tld = random.choice(TRANSLATE_TLDS)
        return f"translate.google{tld}"

    @property
    def translate_url(self):
        url = TRANSLATE_RPC.format(host=self._get_translate_host()) + "?"
        params = {
            "rpcids": RPC_ID,
            "bl": "boq_translate-webserver_20201207.13_p0",
            "soc-app": "1",
            "soc-platform": "1",
            "soc-device": "1",
            "rt": "c",
        }

        return self.format_url(url, params=params)

    async def translate(self, request):
        src_lang, dest_lang = self.denormalize_lang(request.src, request.dest)

        # Form data
        data = {
            "f.req": self._build_rpc_request(request.text, src_lang, dest_lang),
        }

        # Do request
        response = await self.post(self.translate_url, data, self._headers, True, False, False)

        try:
            token_found = False
            square_bracket_counts = [0, 0]
            resp = ""
            data = response.decode("utf-8")

            for line in data.split("\n"):
                token_found = token_found or f'"{RPC_ID}"' in line[:30]
                if not token_found:
                    continue

                is_in_string = False
                for index, char in enumerate(line):
                    if char == '"' and line[max(0, index - 1)] != "\\":
                        is_in_string = not is_in_string
                    if not is_in_string:
                        if char == "[":
                            square_bracket_counts[0] += 1
                        elif char == "]":
                            square_bracket_counts[1] += 1

                resp += line
                if square_bracket_counts[0] == square_bracket_counts[1]:
                    break

            data = json.loads(resp)
            parsed = json.loads(data[0][2])
            translated_parts = None
            translated = None
            try:
                translated_parts = list(
                    map(
                        lambda part: TranslatedPart(
                            part[0] if len(part) > 0 else "", part[1] if len(part) >= 2 else []
                        ),
                        parsed[1][0][0][5],
                    )
                )
            except TypeError:
                translated_parts = [TranslatedPart(parsed[1][0][1][0], [parsed[1][0][0][0], parsed[1][0][1][0]])]

            first_iter = True
            translated = ""
            for part in translated_parts:
                if not part.text.isspace() and not first_iter:
                    translated += " "
                if first_iter:
                    first_iter = False
                translated += part.text

            src = None
            try:
                src = parsed[1][-1][1]
            except (IndexError, TypeError):
                pass

            if not src == src_lang:
                raise UnexpectedError("source language mismatch")

            if src == "auto":
                try:
                    if parsed[0][2] in self.src_languages:
                        src = parsed[0][2]
                except (IndexError, TypeError):
                    pass

            dest = None
            try:
                dest = parsed[1][-1][2]
            except (IndexError, TypeError):
                pass

            if not dest == dest_lang:
                raise UnexpectedError("destination language mismatch")

            origin_pronunciation = None
            try:
                origin_pronunciation = parsed[0][0]
            except (IndexError, TypeError):
                pass

            pronunciation = None
            try:
                pronunciation = parsed[1][0][0][1]
            except (IndexError, TypeError):
                pass

            mistake = None
            try:
                mistake = parsed[0][1][0][0][1]
                # Convert to pango markup
                mistake = mistake.replace("<em>", "<b>").replace("</em>", "</b>")
            except (IndexError, TypeError):
                pass

            return Translation(
                translated,
                request,
                src,
                TranslationMistake(mistake, self._strip_html_tags(mistake)) if mistake else None,
                TranslationPronunciation(origin_pronunciation, pronunciation),
            )

        except Exception as exc:
            raise UnexpectedError from exc

    def _strip_html_tags(self, text: str):
        """Strip html tags"""
        tags_re = re.compile(r"(<!--.*?-->|<[^>]*>)")
        tags_removed = tags_re.sub("", text)
        escaped = html.escape(tags_removed)
        return escaped

    async def speech(self, text, language):
        def get_speech():
            try:
                file = NamedTemporaryFile()
                (lang,) = self.denormalize_lang(language)
                tts = gTTS(text, lang=lang, lang_check=False)
                tts.write_to_fp(file)
                file.seek(0)

                return file
            except Exception:
                raise

        return await self.run_async(get_speech)


class TranslatedPart:
    def __init__(self, text: str, candidates: list[str]):
        self.text = text
        self.candidates = candidates

    def __str__(self):
        return self.text



# Copyright 2024 Mufeed Ali
# Copyright 2024 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from dialect.providers.base import ProviderCapability, ProviderFeature, ProviderLangComparison, Translation
from dialect.providers.errors import APIKeyInvalid, APIKeyRequired, ServiceLimitReached, UnexpectedError
from dialect.providers.soup import SoupProvider

API_V = "v2"


class Provider(SoupProvider):
    name = "deepl"
    prettyname = "DeepL"

    capabilities = ProviderCapability.TRANSLATION
    features = (
        ProviderFeature.DETECTION
        | ProviderFeature.API_KEY
        | ProviderFeature.API_KEY_REQUIRED
        | ProviderFeature.API_KEY_USAGE
    )
    lang_comp = ProviderLangComparison.DEEP

    defaults = {
        "instance_url": "",
        "api_key": "",
        "src_langs": ["en", "fr", "es", "de"],
        "dest_langs": ["fr", "es", "de", "en-US"],
    }

    _api_free = "api-free.deepl.com"
    _api_pro = "api.deepl.com"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.retry_errors = (429,)
        self.chars_limit = 5000

        # DeepL API Free keys can be identified by the suffix ":fx"
        self.api_url = self.__get_api_url(self.api_key)

    def __get_api_url(self, api_key: str) -> str:
        return self._api_free if api_key.endswith(":fx") else self._api_pro

    @property
    def source_lang_url(self):
        return self.format_url(self.api_url, f"/{API_V}/languages", {"type": "source"})

    @property
    def target_lang_url(self):
        return self.format_url(self.api_url, f"/{API_V}/languages", {"type": "target"})

    @property
    def translate_url(self):
        return self.format_url(self.api_url, f"/{API_V}/translate")

    @property
    def usage_url(self):
        return self.format_url(self.api_url, f"/{API_V}/usage")

    @property
    def headers(self):
        return {"Authorization": f"DeepL-Auth-Key {self.api_key}"}

    async def init_trans(self):
        # Get languages
        src_langs = await self.get(self.source_lang_url, self.headers)
        dest_langs = await self.get(self.target_lang_url, self.headers)

        if src_langs and dest_langs and isinstance(src_langs, list) and isinstance(dest_langs, list):
            for lang in src_langs:
                self.add_lang(lang["language"], lang["name"], trans_dest=False)
            for lang in src_langs:
                self.add_lang(lang["language"], lang["name"], trans_src=False)

    async def validate_api_key(self, key):
        api_url = self.__get_api_url(key)
        url = self.format_url(api_url, f"/{API_V}/languages", {"type": "source"})
        headers = {"Authorization": f"DeepL-Auth-Key {key}"}

        try:
            await self.get(url, headers)
            return True
        except (APIKeyInvalid, APIKeyRequired):
            return False
        except Exception:
            raise

    async def translate(self, request):
        src, dest = self.denormalize_lang(request.src, request.dest)

        # Request body
        data = {
            "text": [request.text],
            "target_lang": dest,
        }
        if src != "auto":
            data["source_lang"] = src

        response = await self.post(self.translate_url, data, self.headers)

        # Read translation
        if response and isinstance(response, dict):
            translations: list[dict[str, str]] | None = response.get("translations")
            if translations:
                detected = translations[0].get("detected_source_language")
                translation = Translation(translations[0]["text"], request, detected)
                return translation

        raise UnexpectedError

    async def api_char_usage(self):
        response = await self.get(self.usage_url, self.headers)

        try:
            usage = response.get("character_count")
            limit = response.get("character_limit")

            return usage, limit

        except Exception as exc:
            raise UnexpectedError from exc

    def check_known_errors(self, status, data):
        message = data.get("message", "") if isinstance(data, dict) else ""

        match status:
            case 403:
                if not self.api_key:
                    raise APIKeyRequired(message)
                raise APIKeyInvalid(message)
            case 456:
                raise ServiceLimitReached(message)
            case 429:
                raise UnexpectedError("Too many requests!")

        if status != 200:
            raise UnexpectedError(message)

        if not data:
            raise UnexpectedError
